# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mini-mask creation library."""

from __future__ import annotations

import dataclasses
from typing import Any, Callable, Sequence, Tuple
import numpy as np


# mypy: ignore-errors


class Mask:
	"""A base class for splash attention masks."""

	@property
	def shape(self) -> Tuple[int, ...]:
		raise NotImplementedError

	def __getitem__(self, idx) -> np.ndarray:
		raise NotImplementedError

	def __bool__(self) -> bool:
		raise NotImplementedError(
			"Conversion to bool is unsupported. Could be caused by using logical"
			" instead of bitwise operations on masks."
		)

	def __or__(self, other: "Mask") -> "Mask":
		if self.shape != other.shape:
			raise ValueError(
				f"Invalid shape for other: {other.shape}, expected: {self.shape}"
			)
		return LogicalOr(self, other)

	def __and__(self, other: "Mask") -> "Mask":
		if self.shape != other.shape:
			raise ValueError(
				f"Invalid shape for other: {other.shape}, expected: {self.shape}"
			)
		return LogicalAnd(self, other)


def make_causal_mask(shape: Tuple[int, int], offset: int = 0) -> np.ndarray:
	"""Makes a causal attention mask.

	Args:
	  shape: Shape of the 2-dim mask: (q_seq_len, kv_seq_len).
	  offset: Offset of q start wrt kv. A positive offset shifts the bottom
	    triangle upward, a negative one shifts it downward. A negative offset
	    makes the first 'offset' rows of the attention matrix all 0s which leads
	    to undefined softmax.

	Returns:
	  The causal mask.
	"""
	q_seq_len, kv_seq_len = shape
	q_idx = np.arange(q_seq_len, dtype=np.int32)
	kv_idx = np.arange(kv_seq_len, dtype=np.int32)
	return (q_idx[:, None] + offset >= kv_idx[None, :]).astype(np.bool_)


def make_local_attention_mask(
	shape: Tuple[int, int],
	window_size: Tuple[int | None, int | None],
	*,
	offset: int = 0,
) -> np.ndarray:
	"""Makes a local attention mask."""
	q_seq_len, kv_seq_len = shape
	q_idx = np.arange(q_seq_len, dtype=np.int32)
	kv_idx = np.arange(kv_seq_len, dtype=np.int32)
	mask = np.ones((q_seq_len, kv_seq_len), dtype=np.bool_)
	left, right = window_size
	if left is not None:
		mask = mask & (q_idx[:, None] - left + offset <= kv_idx[None, :])
	if right is not None:
		mask = mask & (q_idx[:, None] + right + offset >= kv_idx[None, :])
	return mask.astype(np.bool_)


def make_random_mask(shape: Tuple[int, int], sparsity: float, seed: int) -> np.ndarray:
	"""Makes a random attention mask."""
	np.random.seed(seed)
	return np.random.binomial(n=1, p=1.0 - sparsity, size=shape).astype(np.bool_)


@dataclasses.dataclass
class LogicalOr(Mask):
	left: Mask
	right: Mask

	def __init__(self, left: Mask, right: Mask):
		if left.shape != right.shape:
			raise ValueError("Masks must have the same shape")
		self.left = left
		self.right = right

	@property
	def shape(self) -> Tuple[int, ...]:
		return self.left.shape

	def __getitem__(self, idx) -> np.ndarray:
		return self.left[idx] | self.right[idx]

	def __hash__(self):
		return hash((type(self),) + (self.left, self.right))


@dataclasses.dataclass
class LogicalAnd(Mask):
	left: Mask
	right: Mask

	def __init__(self, left: Mask, right: Mask):
		if left.shape != right.shape:
			raise ValueError("Masks must have the same shape")
		self.left = left
		self.right = right

	@property
	def shape(self) -> Tuple[int, ...]:
		return self.left.shape

	def __getitem__(self, idx) -> np.ndarray:
		return self.left[idx] & self.right[idx]

	def __hash__(self):
		return hash((type(self),) + (self.left, self.right))


@dataclasses.dataclass
class MultiHeadMask(Mask):
	"""Lazy multihead mask, combines multiple lazy masks one per head."""

	masks: Sequence[Mask]

	def __post_init__(self):
		if not self.masks:
			raise ValueError("Unsupported empty tuple of masks")

		shape = self.masks[0].shape
		for mask in self.masks[1:]:
			if shape != mask.shape:
				raise ValueError(f"Unexpected mask shape, got: {mask.shape}, expected: {shape}")

		if not all(isinstance(mask, Mask) for mask in self.masks):
			raise ValueError("masks should be of type Mask")

		if any(isinstance(mask, MultiHeadMask) for mask in self.masks):
			raise ValueError("Nesting MultiHeadMasks is not supported")

	@property
	def shape(self) -> Tuple[int, ...]:
		return (len(self.masks),) + self.masks[0].shape

	def __getitem__(self, idx) -> np.ndarray:
		if len(idx) != 3:
			raise NotImplementedError(f"Unsupported slice: {idx}")

		head_slice = idx[0]
		if isinstance(head_slice, int):
			assert head_slice >= 0 and head_slice <= len(self.masks)
			return self.masks[head_slice][idx[1:]]
		else:
			slice_masks = [mask[idx[1:]] for mask in self.masks[head_slice]]
			return np.stack(slice_masks)

	def __eq__(self, other: object):
		if not isinstance(other, type(self)):
			return NotImplemented

		return self.masks == other.masks

	def __hash__(self):
		return hash((type(self),) + tuple(hash(mask) for mask in self.masks))


class _ComputableMask(Mask):
	"""Superclass for all masks that can be computed inside the kernel using a callable object.

	Attributes:
	  _shape: Shape of the 2-dim mask: (q_seq_len, kv_seq_len).
	  offset: Offset of q start wrt kv. A positive offset shifts the bottom
	    triangle upward, a negative one shifts it downward. A negative offset
	    makes the first 'offset' rows of the attention matrix all 0s which leads
	    to undefined softmax.
	  q_sequence: Indices of Q sequence.
	    q_sequence is reused across __getitem__ calls which is important for
	    compile-time performance.
	  mask_function: Function used by the SplashAttention kernel to compute the
	    mask rather than loading it.
	"""

	_shape: Tuple[int, int]
	q_sequence: np.ndarray
	mask_function: Callable[..., Any]

	def __init__(
		self,
		shape: Tuple[int, int],
		mask_function: Callable[..., Any],
		shard_count: int = 1,
	):
		self._shape = shape
		self.mask_function = mask_function
		q_seq_len = self.shape[0]

		if q_seq_len % (shard_count * shard_count) != 0:
			raise ValueError(
				f"Shard count squared ({shard_count * shard_count}) must"
				f" divide Q seq_len ({self.shape[0]}) evenly."
			)

		self.q_sequence = np.arange(q_seq_len, dtype=np.int32)

	@property
	def shape(self) -> Tuple[int, ...]:
		return self._shape

	def __getitem__(self, idx) -> np.ndarray:
		if len(idx) != 2:
			raise NotImplementedError(f"Unsupported slice: {idx}")

		q_slice, kv_slice = idx
		if not isinstance(q_slice, slice) or not isinstance(kv_slice, slice):
			raise NotImplementedError(f"Unsupported slice: {idx}")

		q_slice = _fill_slice(q_slice, self.shape[0])
		kv_slice = _fill_slice(kv_slice, self.shape[1])

		rows = self.q_sequence[q_slice]
		cols = np.arange(kv_slice.start, kv_slice.stop)

		return self.mask_function(rows[:, None], cols[None, :])

	def __eq__(self, other: object):
		raise NotImplementedError()

	def __hash__(self):
		raise NotImplementedError()


class CausalMask(_ComputableMask):
	"""Lazy causal mask, prevents the model from attending to future tokens.

	Attributes:
	  offset: Offset of q start wrt kv. A positive offset shifts the bottom
	    triangle upward, a negative one shifts it downward. A negative offset
	    makes the first 'offset' rows of the attention matrix all 0s which leads
	    to undefined softmax.
	"""

	offset: int

	def __init__(
		self,
		shape: Tuple[int, int],
		offset: int = 0,
		shard_count: int = 1,
	):
		self.offset = offset

		def causal_mask_function(q_ids, kv_ids):
			# When evaluating the mask in _process_mask we typically work with numpy
			# array views.
			# Avoid the addition when possible to avoid instantiating an actual array.
			if self.offset == 0:
				return q_ids >= kv_ids
			else:
				return q_ids + self.offset >= kv_ids

		mask_function = causal_mask_function

		super().__init__(
			shape=shape,
			mask_function=mask_function,
			shard_count=shard_count,
		)

	def __eq__(self, other: object):
		if not isinstance(other, type(self)):
			return NotImplemented

		return (
			self.shape == other.shape
			and self.offset == other.offset
			and np.array_equal(self.q_sequence, other.q_sequence)
		)

	def __hash__(self):
		return hash(
			(
				type(self),
				self.shape,
				self.offset,
				self.q_sequence.tobytes() if self.q_sequence is not None else None,
			)
		)


class LocalMask(Mask):
	"""Lazy local mask, prevents model from attending to tokens outside window.

	Attributes:
	  _shape: Shape of the 2-dim mask: (q_seq_len, kv_seq_len).
	  window_size: Size of the two sides of the local window (None identifes no
	    limit for the given side).
	  offset: Offset of q start wrt kv. A positive offset shifts the bottom
	    triangle upward, a negative one shifts it downward. A negative offset
	    makes the first 'offset' rows of the attention matrix all 0s which leads
	    to undefined softmax.
	  _q_sequence: Important for performance.
	"""

	_shape: Tuple[int, int]
	window_size: Tuple[int | None, int | None]
	offset: int
	_q_sequence: np.ndarray | None = None

	def __init__(
		self,
		shape: Tuple[int, int],
		window_size: Tuple[int | None, int | None],
		offset: int,
		shard_count: int = 1,
	):
		self._shape = shape
		self.window_size = window_size
		self.offset = offset

		if self.shape[0] % (shard_count * shard_count) != 0:
			raise ValueError(
				f"Shard count squared ({shard_count * shard_count}) must"
				f" divide Q seq_len ({self.shape[0]}) evenly."
			)

	@property
	def shape(self) -> Tuple[int, int]:
		return self._shape

	def __getitem__(self, idx) -> np.ndarray:
		if len(idx) != 2:
			raise NotImplementedError(f"Unsupported slice: {idx}")
		q_slice, kv_slice = idx
		if not isinstance(q_slice, slice) or not isinstance(kv_slice, slice):
			raise NotImplementedError(f"Unsupported slice: {idx}")

		q_slice = _fill_slice(q_slice, self.shape[0])
		kv_slice = _fill_slice(kv_slice, self.shape[1])

		if self._q_sequence is None:
			rows = np.arange(q_slice.start, q_slice.stop)
		else:
			rows = self._q_sequence[q_slice]

		cols = np.arange(kv_slice.start, kv_slice.stop)

		left_size, right_size = self.window_size

		if left_size is None and right_size is None:
			return np.ones((rows.shape[0], cols.shape[0]), dtype=np.bool_)
		else:
			expanded_cols = cols[None, :]
			if self.offset != 0:
				expanded_rows = rows[:, None] + self.offset
			else:
				expanded_rows = rows[:, None]
			if left_size is not None and right_size is not None:
				return (expanded_rows <= expanded_cols + left_size) & (
					expanded_cols - right_size <= expanded_rows
				)

			elif left_size is not None and right_size is None:
				return expanded_rows <= expanded_cols + left_size
			else:
				assert left_size is None and right_size is not None
				return expanded_cols - right_size <= expanded_rows

	def __eq__(self, other: object):
		if not isinstance(other, type(self)):
			return NotImplemented

		return (
			self.shape == other.shape
			and self.window_size == other.window_size
			and self.offset == other.offset
			and (
				True
				if self._q_sequence is None
				else np.array_equal(self._q_sequence, other._q_sequence)
			)
		)

	def __hash__(self):
		return hash(
			(
				type(self),
				self.shape,
				self.window_size,
				self.offset,
				self._q_sequence.tobytes() if self._q_sequence is not None else None,
			)
		)


@dataclasses.dataclass
class NumpyMask(Mask):
	"""A mask backed by a dense numpy array."""

	array: np.ndarray

	def __post_init__(self):
		if self.array.ndim != 2:
			raise ValueError("Expected a 2-dim array")

		if self.array.dtype != np.bool_:
			raise ValueError("Mask must be a boolean array")

	@property
	def shape(self) -> Tuple[int, ...]:
		return self.array.shape

	def __getitem__(self, idx) -> np.ndarray:
		return self.array[idx]

	def __eq__(self, other: object):
		if not isinstance(other, type(self)):
			return NotImplemented

		return np.array_equal(self.array, other.array, equal_nan=True)

	def __hash__(self):
		return hash((type(self), self.array.tobytes()))


def _fill_slice(inp_slice: slice, size: int) -> slice:
	assert inp_slice.step is None or inp_slice.step == 1
	start = 0 if inp_slice.start is None else inp_slice.start
	stop = size if inp_slice.stop is None else inp_slice.stop
	assert start >= 0
	assert stop <= size
	return slice(start, stop, None)


@dataclasses.dataclass(frozen=True)
class FullMask(Mask):
	"""Lazy full mask, allows all tokens to attend to all other tokens."""

	_shape: tuple[int, int]

	def __post_init__(self):
		if not isinstance(self.shape, tuple):
			raise ValueError(f"Unsupported shape type: {type(self.shape)}")

	@property
	def shape(self) -> Tuple[int, ...]:
		return self._shape

	def __getitem__(self, idx) -> np.ndarray:
		if len(idx) != 2:
			raise NotImplementedError(f"Unsupported slice: {idx}")
		i, j = idx
		if not isinstance(i, slice) or not isinstance(j, slice):
			raise NotImplementedError(f"Unsupported slice: {idx}")
		i = _fill_slice(i, self.shape[0])
		j = _fill_slice(j, self.shape[1])
		return np.ones((i.stop - i.start, j.stop - j.start), dtype=np.bool_)

	def __eq__(self, other: object):
		if not isinstance(other, type(self)):
			return NotImplemented

		return self.shape == other.shape

	def __hash__(self):
		return hash((type(self), self.shape))



# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mini-mask creation library."""

from __future__ import annotations

import collections
import functools
from typing import Callable, Dict, List, NamedTuple, Set, Tuple
from jax import util as jax_util
from fjformer.pallas_operations.tpu.splash_attention import (
	splash_attention_mask as mask_lib,
)
import numpy as np


# mypy: ignore-errors


# Logic for processing NumPy masks for kernels
class MaskInfo(NamedTuple):
	"""Contains runtime masking information for the Splash attention kernel.

	The arrays data_next, mask_next and block_mask are placed in TPU
	scalar-memory. This is a scarse resource so the mask creation logic attempts
	to shrink the data-type of these arrays to the smallest possible one.
	This can be: np.int32, np.int16 or np.int8.

	For the arrays data_next, mask_next and block_mask the size of the first
	dimension can be one of the two following values: num_head or
	num_head_shards.
	The first dimension has size:
	* num_head_shards when there is only one unique mask for each head in a shard.
	In this case the three arrays are broadcasted to all the heads in the shard.
	* num_heads when there is more than one unique mask for each head in the
	shard.

	Attributes:
	  data_next: An integer[num_heads_or_shards, num_q_blocks, num_kv_blocks]
	    NumPy array where each entry contains the next `kv` block index to
	    prefetch.
	  mask_next: An integer[num_heads_or_shards, num_q_blocks, num_kv_blocks]
	    NumPy array where each entry contains the next mask block index in
	    `partial_mask_blocks` to prefetch.
	  block_mask: An integer[num_heads_or_shards, num_q_blocks, num_kv_blocks]
	    NumPy array whose entries can be 0, 1 or 2. An entry of 0 indicates that
	    the corresponding block in the full mask was all zeros. An entry of 1
	    indicates that the corresponding block in the full mask contained both
	    zeros and ones. An entry of 2 indicates the corresponding block was
	    entirely ones.
	  partial_mask_blocks: A i32[num_partial_blocks, block_q, block_kv] NumPy
	    array that contains the blocks of the original mask that contained both
	    zeros and ones. The entries in `mask_next` point to indices in the first
	    axis of this array.
	  q_sequence: A i32[q_sequence_length] NumPy array. When using causal masking,
	    this contains the list of indices that correspond to q tokens. For plain
	    causal this is just np.arange(q_sequence_length).
	"""

	data_next: np.ndarray | None
	mask_next: np.ndarray | None
	block_mask: np.ndarray | None
	partial_mask_blocks: np.ndarray | None
	q_sequence: np.ndarray | None


def _downcast_to_small_type(array: np.ndarray) -> np.ndarray:
	"""Downcast numpy array.

	If possible, downcast the data-type of the input array to the smallest numpy
	type (among np.int16 and np.int8) that fits the content of the array.

	Args:
	  array: the array to downcast

	Returns:
	  The downcasted array.

	Raises:
	  ValueError: if the input array is not np.int32 or if its elements are not
	  all positive.
	"""

	if array.dtype != np.int32:
		raise ValueError("Expected int32 input.")

	if not np.all(array >= 0):
		raise ValueError("Expected non-negative array.")

	if array.size == 0:
		return array

	max_value = np.max(array)

	if max_value <= np.iinfo(np.int8).max:
		return array.astype(np.int8)
	elif max_value <= np.iinfo(np.int16).max:
		return array.astype(np.int16)
	else:
		return array.astype(np.int32)


def _check_mask(mask: mask_lib.Mask) -> None:
	"""Check that the given mask is valid.

	A row of all zeros along the kv dimension would result in a division by zero
	when computing the softmax. This function is meant to protect against that
	case.

	Args:
	  mask: the mask to check.

	Raises:
	  ValueError: the mask is invalid.
	"""

	assert len(mask.shape) == 2

	exception_message = (
		"Some rows of the mask (along the kv dimension) are all zeros.\nThis is"
		" would result in a division by zero when computing the attention"
		" softmax."
	)

	is_row_non_zero = np.zeros(mask.shape[0], dtype=np.bool_)
	for col in range(mask.shape[1]):
		# Mask only supports slice indices.
		is_row_non_zero = np.logical_or(
			is_row_non_zero,
			mask[(slice(0, mask.shape[0]), slice(col, col + 1))][:, 0],
		)
	if not is_row_non_zero.all():
		raise ValueError(exception_message)


class _HashableNDArray:
	"""Helper to make a numpy array hashable: can be added associative containers.

	Attributes:
	  array: The underlying numpy array.
	"""

	array: np.ndarray

	def __init__(self, array: np.ndarray):
		self.array = array

	def __hash__(self):
		return hash(self.array.tobytes())

	def __eq__(self, other: object) -> bool:
		if not isinstance(other, _HashableNDArray):
			return NotImplemented
		return np.array_equal(self.array, other.array, equal_nan=True)


def _get_mask_info_for_shard(
	output_shape: Tuple[int, int, int],
	has_mask_next: bool,
	mask: mask_lib.MultiHeadMask,
	block_shape: Tuple[int, int],
	coords_to_partial_mask_block_index: Dict[Tuple[int, int, int], int],
	masks_per_head_shard: int,
	head_start: int,
	num_heads: int,
	q_seq_start: int,
	q_seq_shard_size: int,
	blocked_q_seq_start: int,
	is_dkv: bool,
) -> Tuple[np.ndarray, np.ndarray | None]:
	"""Process a slice of the mask to compute data_next and mask_next.

	Args:
	  output_shape: The shape of the data_next and mask_next to return
	  has_mask_next: Whether mask_next should be constructed. If False None is
	    returned for mask_next.
	  mask: The full mask to be sliced according to the head and sequence ranges
	  block_shape: Shape of the Pallas grid block.
	  coords_to_partial_mask_block_index: Mapping between the pallas launch grid
	    coordinates and the index of the corresponding block in partial mask block
	    list.
	  masks_per_head_shard: Number of masks per head shards
	  head_start: First head of the current shard.
	  num_heads: Number of heads in the shard.
	  q_seq_start: Start index along the Q sequence for the current shard (in
	    number of tokens).
	  q_seq_shard_size: Number of tokens along the Q sequence for the current
	    shard.
	  blocked_q_seq_start: Start index along the Q sequence for the current shard
	    (in number of grid blocks)
	  is_dkv: True if we are processing the dKV mask

	Returns:
	  Slice of data_next and mask_next (if required) that correspond to the
	  current mask slice.
	"""
	_, _, kv_seq_len = mask.shape

	q_block_size, kv_block_size = block_shape
	q_block_count, q_mod = divmod(q_seq_shard_size, q_block_size)
	kv_block_count, kv_mod = divmod(kv_seq_len, kv_block_size)

	assert q_mod == 0
	assert kv_mod == 0

	blocked_shape = (
		(kv_block_count, num_heads, q_block_count)
		if is_dkv
		else (num_heads, q_block_count, kv_block_count)
	)

	data_coords = []
	mask_coords = []
	for idx in np.ndindex(blocked_shape):
		if is_dkv:
			kv_index, h_index, q_index = idx
		else:
			h_index, q_index, kv_index = idx

		h_index = h_index if masks_per_head_shard == 1 else head_start + h_index

		chunk = mask[
			(
				h_index,
				slice(
					q_seq_start + q_index * q_block_size,
					q_seq_start + (q_index + 1) * q_block_size,
				),
				slice(kv_index * kv_block_size, (kv_index + 1) * kv_block_size),
			)
		]
		if chunk.any():
			data_coords.append(idx)
			if not chunk.all():
				mask_coords.append(idx)

	# Initialize the output arrays.
	mask_next = None
	if has_mask_next:
		mask_next = np.zeros(output_shape, dtype=np.int32)
	data_next = np.zeros(output_shape, dtype=np.int32)

	# If the mask is completelly zero'd out return freshly initialized outputs.
	if not data_coords:
		return data_next, mask_next

	data_coords_iter = iter(data_coords)
	first_j = coord_j = next(data_coords_iter)
	if mask_next is not None and mask_coords:
		mask_coords_iter = iter(mask_coords)
		first_m = coord_m = next(mask_coords_iter)
	else:
		first_m, coord_m, mask_coords_iter = None, None, None

	for idx in np.ndindex(blocked_shape):
		if is_dkv:
			kv_index, h_index, q_index = idx
			chunk_idx: tuple[int, ...] = (h_index, q_index, kv_index)
			data_dim = 2
		else:
			chunk_idx = idx
			data_dim = 2

		is_next = idx > coord_j

		if is_next:
			try:
				coord_j = next(data_coords_iter)
			except StopIteration:
				coord_j = first_j
		data_next[chunk_idx] = coord_j[data_dim]

		if mask_next is not None and mask_coords:
			assert coord_m is not None
			is_next_mask = idx > coord_m
			if is_next_mask:
				try:
					coord_m = next(mask_coords_iter)  # type: ignore
				except StopIteration:
					coord_m = first_m

			if is_dkv:
				assert coord_m is not None
				coord_m_global = (
					coord_m[1] + head_start,
					coord_m[2] + blocked_q_seq_start,
					coord_m[0],
				)
			else:
				assert coord_m is not None
				coord_m_global = (
					coord_m[0] + head_start,
					coord_m[1] + blocked_q_seq_start,
					coord_m[2],
				)

			mask_next[chunk_idx] = coords_to_partial_mask_block_index[coord_m_global]

	return data_next, mask_next


# When used in a transformer network with multiple layers, the SplashAttention
# kernel is created serveral times with the same mask. Cache MaskInfo to avoid
# blowing up compile times. Ideally the size of the cache should be determined
# by the client.
@functools.lru_cache(maxsize=12)
def _process_mask(
	mask: mask_lib.MultiHeadMask,  # [num_heads, q_seq_len, kv_seq_len]
	block_shape: Tuple[int, int],
	is_dkv: bool,
	*,
	downcast_smem_data: bool = True,
	head_shards: int = 1,
	q_seq_shards: int = 1,
	shrink_grid: bool = True,
) -> tuple[MaskInfo, jax_util.HashableFunction | None]:
	"""Transform a dense mask into a sparse representation.

	The number of head and Q sequence shards are needed to create a MaskInfo
	object that is partitionable (with shmap or PartIR) along these two dimension.
	In particular for dKV MaskInfo, for each shard the indices of in the data_next
	array are relative to the current shard.
	The fwd and dQ MaskInfo objects do not change when sharding along the head or
	Q dimensions, they would be different if we were to shard along the KV
	dimension, but the kernel does not support that.

	Args:
	  mask: Dense mask to process.
	  block_shape: Shape of the Pallas grid block.
	  is_dkv: True if we are processing the dKV mask
	  downcast_smem_data: If True, downcast the scalar-memory data of MaskInfo to
	    a data type smaller than np.int32 (if possible).
	  head_shards: Number of head shards of the mesh in which the kernel is
	    launched.
	  q_seq_shards: Number of Q sequence shards of the mesh in which the kernel is
	    launched.
	  shrink_grid: Whether or not we should apply the grid shrinking optimization.

	Returns:
	  `MaskInfo`, a sparse representation of the dense mask.
	  `MaskCallable`: a callable that, given in input Q and KV indices, returns
	    the value of the mask at those coordinates.

	Raises:
	  ValueError: if the input mask is invalid or the block sizes are not
	  compatible with the mask sizes.
	"""

	if len(mask.shape) != 3:
		raise ValueError(f"Expected a 3-dim mask, instead got: {mask.shape=}")

	head_count, q_seq_len, kv_seq_len = mask.shape
	q_block_size, kv_block_size = block_shape
	q_blocks_count, q_mod = divmod(q_seq_len, q_block_size)
	kv_blocks_count, kv_mod = divmod(kv_seq_len, kv_block_size)

	if q_mod != 0:
		raise ValueError(f"{q_block_size=} should divide {q_seq_len=}.")
	if kv_mod != 0:
		raise ValueError(f"{kv_block_size=} should divide {kv_seq_len=}.")

	q_seq_len_per_shard, mod = divmod(q_seq_len, q_seq_shards)
	if mod != 0:
		raise ValueError(f"{q_seq_shards=} should divide {q_seq_len=}.")

	q_blocks_per_shard, mod = divmod(q_seq_len_per_shard, q_block_size)
	if mod != 0:
		raise ValueError(f"{q_block_size=} should divide {q_seq_len_per_shard=}.")

	heads_per_shard, mod = divmod(head_count, head_shards)
	if mod != 0:
		raise ValueError(f"{head_shards=} should divide {head_count=}.")

	first_mask_size = mask.masks[0].shape
	for h in range(head_count):
		if mask.masks[h].shape != first_mask_size:
			raise ValueError(
				f"First head mask has shape {first_mask_size}, but head mask {h} has"
				f" shape {mask.masks[h].shape}. All head masks must have the same"
				" shape."
			)

	# Uniquify the masks.
	# Create a collection of the unique head masks in the input multi-head mask.
	# This avoids processing the same mask multiple times and it enables
	# deduplicating the partial mask blocks information.
	# Assign a unique ID to every unique mask.

	def assign_unique_ids(objects):
		id_map = collections.defaultdict(lambda: len(id_map))
		return {obj: id_map[obj] for obj in objects}

	unique_masks_dict: Dict[mask_lib.Mask, int] = assign_unique_ids(
		head_mask for head_mask in mask.masks
	)

	# Build a mapping of heads to unique masks and masks to unique masks.

	head_to_mask_id: List[int] = [0] * head_count
	head_shard_to_mask_ids: List[Set[int]] = [set() for _ in range(head_shards)]
	mask_id_to_heads: List[List[int]] = [[] for _ in range(len(unique_masks_dict))]
	mask_id_to_head_shards: List[Set[int]] = [
		set() for _ in range(len(unique_masks_dict))
	]

	for head in range(head_count):
		mask_id = unique_masks_dict[mask.masks[head]]
		head_to_mask_id[head] = mask_id
		head_shard = head // heads_per_shard
		head_shard_to_mask_ids[head_shard].add(mask_id)
		mask_id_to_heads[mask_id].append(head)
		mask_id_to_head_shards[mask_id].add(head_shard)

	# If we have at most one unique mask per each head shard, then we can brodcast
	# the mask to all the heads in the shard. This is the common case.
	# If we have more than one mask in each head shard, then the optimization
	# cannot kick in and we use one mask for each head.
	# heads to masks. This will likely introduce an additional field in the
	# MaskInfo class and runtime overhead to perform an indirect lookup. Since
	# having multiple masks per head-shard is not a common case we leave this for
	# future work.
	max_masks_per_head_shard = max([len(x) for x in head_shard_to_mask_ids])
	masks_per_head_shard = 1 if max_masks_per_head_shard == 1 else heads_per_shard

	unique_masks = [
		pair[0] for pair in sorted(unique_masks_dict.items(), key=lambda x: x[1])
	]

	partial_mask_block_ids: Dict[_HashableNDArray, int] = collections.defaultdict(
		lambda: len(partial_mask_block_ids)
	)
	block_id_to_block_coords: Dict[int, List[Tuple[int, ...]]] = collections.defaultdict(
		list
	)

	# Shape of the block_mask, mask_next and data_next fields of MaskInfo.
	block_mask_shape = (
		head_shards if masks_per_head_shard == 1 else head_count,
		q_blocks_count,
		kv_blocks_count,
	)

	block_mask = np.zeros(block_mask_shape, dtype=np.int32)

	def set_block_mask(mask_id: int, q_index: int, kv_index: int, value: int):
		if masks_per_head_shard == 1:
			for shard_index in mask_id_to_head_shards[mask_id]:
				block_mask[shard_index, q_index, kv_index] = value
		else:
			for head_index in mask_id_to_heads[mask_id]:
				block_mask[head_index, q_index, kv_index] = value

	q_sequence = None
	mask_function = None
	if len(unique_masks) == 1:
		unique_mask = unique_masks[0]
		# The mask object either define q_sequence and mask_function or none of
		# them.
		assert hasattr(unique_mask, "q_sequence") == hasattr(unique_mask, "mask_function")

		# If the mask object defines a q_sequence and a mask_function, then make use
		# of these in the kernel rather. This is preferable over loading the mask
		# from memory. When using a mask_function, then mask_next and
		# partial_mask_blocks are left undefined and not used in the kernel.
		if hasattr(unique_mask, "q_sequence") and hasattr(unique_mask, "mask_function"):
			q_sequence = unique_mask.q_sequence
			mask_function = unique_mask.mask_function

	# Identify the partial mask blocks and the value of the block mask for each
	# block.
	# The partial mask blocks are "global", meaning it is a collection of such
	# blocks across all the heads. Partial mask blocks are uniquified.
	# When partitioning, all partial mask blocks are replicated across shards.
	# As a future extension we could assign to each head shard only the relevant
	# partial blocks. This would reduce the amount of data each shard needs
	# from: num_unique_partial_block_blocks to
	# max(unique_partial_blocks_per_shard)
	for mask_id, unique_mask in enumerate(unique_masks):
		for coords in np.ndindex((q_blocks_count, kv_blocks_count)):
			(q_index, kv_index) = coords
			chunk = unique_mask[
				(
					slice(q_index * q_block_size, (q_index + 1) * q_block_size),
					slice(kv_index * kv_block_size, (kv_index + 1) * kv_block_size),
				)
			]
			has_nonzero = chunk.any()
			if has_nonzero:
				all_nonzero = chunk.all()
				if not all_nonzero:
					set_block_mask(mask_id, q_index, kv_index, 1)
					partial_mask_block_id = partial_mask_block_ids[_HashableNDArray(chunk)]
					for head_index in mask_id_to_heads[mask_id]:
						block_id_to_block_coords[partial_mask_block_id].append(
							(head_index,) + coords
						)
				else:
					set_block_mask(mask_id, q_index, kv_index, 2)

	unique_partial_mask_blocks = [
		pair[0] for pair in sorted(partial_mask_block_ids.items(), key=lambda x: x[1])
	]

	# For each position in the in q, kv grid that contains a partial mask block
	# record the index of its partial mask block.
	coords_to_partial_mask_block_index = {}
	for partial_mask_block_id, coords in block_id_to_block_coords.items():
		for coo in coords:
			coords_to_partial_mask_block_index[coo] = partial_mask_block_id

	partial_mask_blocks = None
	has_mask_next = False
	if len(unique_partial_mask_blocks) == 1:
		partial_mask_blocks = [x.array for x in unique_partial_mask_blocks]
		partial_mask_blocks = partial_mask_blocks[0][None].astype(np.int32)
		has_mask_next = True
	elif len(unique_partial_mask_blocks) > 1:
		partial_mask_blocks = [x.array for x in unique_partial_mask_blocks]
		partial_mask_blocks = np.stack(partial_mask_blocks, axis=0).astype(np.int32)
		has_mask_next = True
	if is_dkv and partial_mask_blocks is not None:
		partial_mask_blocks = np.swapaxes(partial_mask_blocks, -1, -2)

	all_head_shards_identical = all(
		head_shard_to_mask_ids[0] == x and len(x) == 1 for x in head_shard_to_mask_ids
	)
	# When all the head shards are identical, we can process only one and
	# broadcast the result across all the head shards, this avoids redundant work.
	shards_to_process = 1 if all_head_shards_identical else head_shards

	# Iterate over the shards.
	# Work on a fraction of the mask at the time to compute the mask and the data
	# indices. This is needed to compute the correct data indices, which are
	# relative to the current slice of the mask.

	q_sequence_axis = 1
	head_axis = 0

	# Collect mask_info shards along the head dimension, concatentate (or
	# broadcast) them after the loop.
	data_next_per_head_list, mask_next_per_head_list = [], []
	for head_shard in range(shards_to_process):
		# Collect mask_info shards along the q_sequence dimension, concatenate them
		# after the loop.
		data_next_sequence_slices, mask_next_sequence_slices = [], []
		for q_seq_len_shard in range(q_seq_shards):
			head_start = head_shard * heads_per_shard
			q_seq_len_shard_size = q_blocks_per_shard * q_block_size
			q_seq_len_start = q_seq_len_shard * q_seq_len_shard_size

			blocked_q_seq_len_start = q_seq_len_shard * q_blocks_per_shard
			blocked_q_seq_len_slice = slice(
				blocked_q_seq_len_start,
				(q_seq_len_shard + 1) * q_blocks_per_shard,
			)

			if masks_per_head_shard == 1:
				unique_mask = unique_masks[head_to_mask_id[head_start]]
				unique_mask = mask_lib.MultiHeadMask((unique_mask,))
				current_mask = unique_mask
				mask_head_slice = slice(head_shard, head_shard + 1)
			else:
				current_mask = mask
				mask_head_slice = slice(head_start, (head_shard + 1) * heads_per_shard)

			# The current iteration of the loop processes a slice of the mask info
			# tensors of this shape:
			mask_info_slice_shape = (
				mask_head_slice.stop - mask_head_slice.start,
				blocked_q_seq_len_slice.stop - blocked_q_seq_len_slice.start,
				kv_blocks_count,
			)
			# Generate data_next and mask_next for the current slice.
			data_next_slice, mask_next_slice = _get_mask_info_for_shard(
				output_shape=mask_info_slice_shape,
				has_mask_next=has_mask_next,
				mask=current_mask,
				block_shape=block_shape,
				coords_to_partial_mask_block_index=coords_to_partial_mask_block_index,
				head_start=head_start,
				masks_per_head_shard=masks_per_head_shard,
				num_heads=1 if masks_per_head_shard == 1 else heads_per_shard,
				q_seq_start=q_seq_len_start,
				q_seq_shard_size=q_seq_len_shard_size,
				blocked_q_seq_start=blocked_q_seq_len_start,
				is_dkv=is_dkv,
			)
			data_next_sequence_slices.append(data_next_slice)
			mask_next_sequence_slices.append(mask_next_slice)

		# Concatenate the sequence shards.
		data_next_per_head = np.concatenate(data_next_sequence_slices, axis=q_sequence_axis)
		data_next_per_head_list.append(data_next_per_head)
		if has_mask_next:
			mask_next_per_head = np.concatenate(
				mask_next_sequence_slices, axis=q_sequence_axis
			)
			mask_next_per_head_list.append(mask_next_per_head)

	# Concatenate (or broadcast) the head shards.
	mask_next = None
	if all_head_shards_identical:
		assert len(data_next_per_head_list) == 1
		data_next_shard = data_next_per_head_list[0]
		assert data_next_shard.shape == (1, q_blocks_count, kv_blocks_count)
		data_next = np.broadcast_to(
			data_next_shard,
			(head_shards, q_blocks_count, kv_blocks_count),
		)
		if has_mask_next:
			assert len(mask_next_per_head_list) == 1
			mask_next_shard = mask_next_per_head_list[0]
			assert mask_next_shard.shape == (1, q_blocks_count, kv_blocks_count)
			mask_next = np.broadcast_to(
				mask_next_shard,
				(head_shards, q_blocks_count, kv_blocks_count),
			)
	else:
		data_next = np.concatenate(data_next_per_head_list, axis=head_axis)
		if has_mask_next:
			mask_next = np.concatenate(mask_next_per_head_list, axis=head_axis)

	# Shrink the width of the mask info when possible.
	if shrink_grid and block_mask.shape[0] == head_shards and len(unique_masks) == 1:
		rows_per_q_shard = block_mask.shape[1] // q_seq_shards
		block_mask_shards = []
		data_next_shards = []
		mask_next_shards = []

		# Slice the mask info arrays along the Q dimension. This simulates Q
		# sharding and enabled shrinking each Q shard of the MaskInfo arrays
		# independently.
		for q_seq_len_shard in range(q_seq_shards):
			rows = slice(
				q_seq_len_shard * rows_per_q_shard,
				(q_seq_len_shard + 1) * rows_per_q_shard,
			)
			current_block_mask = block_mask[:, rows, :]
			current_data_next = data_next[:, rows, :]
			current_mask_next = mask_next[:, rows, :] if mask_next is not None else None

			shrink_function = _shrink_mask_info_dkv if is_dkv else _shrink_mask_info

			current_block_mask, current_data_next, current_mask_next = shrink_function(
				block_mask=current_block_mask,
				data_next=current_data_next,
				mask_next=current_mask_next,
				head_shards=head_shards,
			)

			assert current_block_mask.size > 0
			assert current_data_next.size > 0
			assert current_mask_next is None or current_mask_next.size > 0
			assert current_block_mask.shape == current_data_next.shape
			assert (
				current_mask_next is None or current_block_mask.shape == current_mask_next.shape
			)

			block_mask_shards.append(current_block_mask)
			data_next_shards.append(current_data_next)
			mask_next_shards.append(current_mask_next)

		if q_seq_shards == 1:
			block_mask = block_mask_shards[0]
			data_next = data_next_shards[0]
			mask_next = mask_next_shards[0]
		else:
			# Since shrinking happens independently for each shard along Q we might
			# end up with uneven mask info shards. Insert padding where necessary to
			# maintain the SPMD paradigm.
			padding_axis = 1 if is_dkv else 2

			max_size = max([x.shape[padding_axis] for x in block_mask_shards])
			padded_block_mask_shards = []
			padded_data_next_shards = []
			padded_mask_next_shards = []
			assert len(block_mask_shards) == len(data_next_shards) == len(mask_next_shards)
			for (
				current_block_mask,
				current_data_next,
				current_mask_next,
			) in zip(block_mask_shards, data_next_shards, mask_next_shards):
				# For dKV shrinking happens along axis Q (the rows of MaskInfo), for
				# fwd and dQ shrinking happens along axis KV (the columns of MaskInfo).
				if is_dkv:
					pad_width = (
						(0, 0),
						(0, max_size - current_block_mask.shape[padding_axis]),
						(0, 0),
					)
				else:
					pad_width = (
						(0, 0),
						(0, 0),
						(0, max_size - current_block_mask.shape[padding_axis]),
					)

				padded_block_mask_shards.append(
					np.pad(current_block_mask, pad_width=pad_width, constant_values=0)
				)
				# By padding data_next and mask_next with 'edge' policy we avoid
				# fetching new blocks thanks to 'immediate revisiting'.
				padded_data_next_shards.append(
					np.pad(current_data_next, pad_width=pad_width, mode="edge")
				)
				if current_mask_next is not None:
					padded_mask_next_shards.append(
						np.pad(current_mask_next, pad_width=pad_width, mode="edge")
					)

			block_mask = np.concatenate(padded_block_mask_shards, axis=1)
			data_next = np.concatenate(padded_data_next_shards, axis=1)
			mask_next = np.concatenate(padded_mask_next_shards, axis=1)

	if downcast_smem_data:
		data_next = _downcast_to_small_type(data_next)
		block_mask = _downcast_to_small_type(block_mask)
		if mask_next is not None:
			mask_next = _downcast_to_small_type(mask_next)

	assert (mask_function is not None) == (q_sequence is not None)
	# When the mask can be computed inside the kernel with a mask_function,
	# there is no need to load it from memory. So mask_next and
	# partial_mask_blocks are unused.
	return (
		MaskInfo(
			data_next=data_next,
			mask_next=mask_next if mask_function is None else None,
			block_mask=block_mask,
			partial_mask_blocks=partial_mask_blocks if mask_function is None else None,
			q_sequence=q_sequence,
		),
		mask_function,
	)


def _shrink_mask_info(
	*,
	block_mask: np.ndarray,
	data_next: np.ndarray,
	mask_next: np.ndarray,
	head_shards: int,
):
	assert block_mask.ndim == 3
	assert data_next.ndim == 3
	assert mask_next is None or mask_next.ndim == 3

	head_block_mask = block_mask[0]

	grouped_non_zero_cols = []
	# Group non-zero columns based on which row they belong to.
	for row_index in range(head_block_mask.shape[0]):
		head_block_mask_row = head_block_mask[row_index, :]
		non_zero_cols = np.nonzero(head_block_mask_row)[0]
		grouped_non_zero_cols.append(non_zero_cols)

	# Pad each row in the non-zero indices to match the width of the longest
	# row. This avoids having jagged rows.
	max_non_zero_cols = max([len(x) for x in grouped_non_zero_cols])
	padded_non_zero_cols = []
	padding = -1
	for row in grouped_non_zero_cols:
		padded_non_zero_cols.append(
			np.pad(
				row,
				pad_width=(0, max_non_zero_cols - row.shape[0]),
				constant_values=padding,
			)
		)

	padded_non_zero_cols = np.stack(padded_non_zero_cols, axis=0)

	assert padded_non_zero_cols.shape[0] == block_mask.shape[1], (
		padded_non_zero_cols.shape,
		block_mask.shape,
	)

	# For each row of array, select the columns indices in padded_non_zero_cols,
	# ignore padding.
	def select_cols(array):
		assert array.ndim == 2
		assert padded_non_zero_cols.ndim == 2
		assert array.shape[0] == padded_non_zero_cols.shape[0]
		assert array.shape[1] >= padded_non_zero_cols.shape[1]
		selected_rows = []
		for row in range(array.shape[0]):
			col = padded_non_zero_cols[row]
			selected = array[row][col]
			selected = np.where(col != padding, selected, 0)
			selected_rows.append(selected)

		return np.stack(selected_rows, axis=0)

	return _slice_mask_info(
		block_mask=block_mask,
		data_next=data_next,
		mask_next=mask_next,
		head_shards=head_shards,
		slice_function=select_cols,
	)


def _shrink_mask_info_dkv(
	*,
	block_mask: np.ndarray,
	data_next: np.ndarray,
	mask_next: np.ndarray,
	head_shards: int,
):
	assert block_mask.ndim == 3
	assert data_next.ndim == 3
	assert mask_next is None or mask_next.ndim == 3

	head_block_mask = block_mask[0]
	grouped_non_zero_rows = []
	# Group non-zero rows based on which column they belong to.
	for col_index in range(head_block_mask.shape[1]):
		col = head_block_mask[:, col_index]
		non_zero_rows = np.nonzero(col)[0]
		grouped_non_zero_rows.append(non_zero_rows)

	# Pad each col in the non-zero indices to match the height of the longest
	# col. This avoids having jagged cols.
	max_non_zero_rows = max([len(x) for x in grouped_non_zero_rows])
	padded_non_zero_rows = []
	padding = -1
	for col in grouped_non_zero_rows:
		padded_non_zero_rows.append(
			np.pad(
				col,
				pad_width=(max_non_zero_rows - col.shape[0], 0),
				constant_values=padding,
			)
		)

	padded_non_zero_rows = np.stack(padded_non_zero_rows, axis=1)

	assert padded_non_zero_rows.shape[1] == block_mask.shape[2], (
		padded_non_zero_rows.shape,
		block_mask.shape,
	)

	# For each col of array, select the rows indices in padded_non_zero_rows,
	# ignore padding.
	def select_rows(array):
		assert array.ndim == 2
		assert padded_non_zero_rows.ndim == 2
		assert array.shape[1] == padded_non_zero_rows.shape[1]
		assert array.shape[0] >= padded_non_zero_rows.shape[0]
		selected_cols = []
		for col in range(array.shape[1]):
			row = padded_non_zero_rows[:, col]
			selected = array[:, col][row]
			selected = np.where(row != padding, selected, 0)
			selected_cols.append(selected)

		return np.stack(selected_cols, axis=1)

	return _slice_mask_info(
		block_mask=block_mask,
		data_next=data_next,
		mask_next=mask_next,
		head_shards=head_shards,
		slice_function=select_rows,
	)


def _slice_mask_info(
	*,
	block_mask: np.ndarray,
	data_next: np.ndarray,
	mask_next: np.ndarray,
	head_shards: int,
	slice_function: Callable[[np.ndarray], np.ndarray],
):
	new_block_mask = []
	new_data_next = []
	new_mask_next = []
	for head_shard in range(head_shards):
		head_block_mask = block_mask[head_shard]
		head_block_mask = slice_function(head_block_mask)
		new_block_mask.append(head_block_mask)

		head_data_next = data_next[head_shard]
		head_data_next = slice_function(head_data_next)
		new_data_next.append(head_data_next)

		if mask_next is not None:
			head_mask_next = mask_next[head_shard]
			head_mask_next = slice_function(head_mask_next)
			new_mask_next.append(head_mask_next)

	block_mask = np.stack(new_block_mask, axis=0)
	data_next = np.stack(new_data_next, axis=0)
	if mask_next is not None:
		mask_next = np.stack(new_mask_next, axis=0)

	return block_mask, data_next, mask_next


process_mask = functools.partial(_process_mask, is_dkv=False)
process_mask_dkv = functools.partial(_process_mask, is_dkv=True)



# jax.experimental.pallas.ops
from __future__ import annotations

import functools
from typing import Any

import jax
import jax.numpy as jnp
import numpy as np
from jax import lax
from jax.experimental import pallas as pl

DEFAULT_MASK_VALUE = -0.7 * float(np.finfo(np.dtype("float32")).max)


def mha_forward_kernel(
	q_ref,
	k_ref,
	v_ref,  # Input arrays
	segment_ids_ref: jax.Array | None,  # segment_id arrays
	o_ref: Any,  # Output
	*residual_refs: Any,  # Residual outputs
	num_heads: int,
	sm_scale: float,
	causal: bool,
	block_q: int,
	block_d: int,
	block_k: int,
):
	seq_len = q_ref.shape[0]
	start_q = pl.program_id(0)

	# o is the buffer where we accumulate the output on sram.
	# m_i and l_i (see FlashAttention paper) are updated during the k,v loop.
	m_i = jnp.zeros(block_q, dtype=jnp.float32) - float("inf")
	l_i = jnp.zeros(block_q, dtype=jnp.float32)
	# acc is the buffer where we accumulate the output on sram.
	o = jnp.zeros((block_q, block_d), dtype=jnp.float32)

	# Load q: it will stay in L1 throughout. Indices form a matrix because we
	# read, compute, and write all in 2d chunks. 1 element ~= 1 CUDA thread index.
	# q tile has shape [block_q, block_d], block_d == head_dim.
	curr_q_slice = pl.dslice(start_q * block_q, block_q)
	q = pl.load(q_ref, (curr_q_slice, pl.dslice(None)))
	q_segment_ids = (
		None if segment_ids_ref is None else pl.load(segment_ids_ref, (curr_q_slice,))
	)

	# In FlashAttention algorithm 1 there are 2 loops: slow over tiles of kv (size
	# (Bc == block_k here), and fast over blocks of q (size Br == block_q here).
	# Here we only loop over blocks of kv to process entire seq_len, the loop over
	# blocks of q is carried out by the grid.
	def body(start_k, carry):
		o_prev, m_prev, l_prev = carry
		curr_k_slice = pl.dslice(start_k * block_k, block_k)

		k = pl.load(k_ref, (curr_k_slice, slice(None)))
		kv_segment_ids = (
			None if segment_ids_ref is None else pl.load(segment_ids_ref, (curr_k_slice,))
		)
		qk = pl.dot(q, k.T)  # [block_q, block_k]
		if sm_scale != 1.0:
			qk *= sm_scale  # [block_q, block_k]

		# Avoids Triton crash.
		# if num_heads > 2:
		#   qk = qk.astype(q_ref.dtype)
		#   qk = qk.astype(jnp.float32)

		if causal or segment_ids_ref is not None:
			mask = None
			if segment_ids_ref is not None:
				mask = segment_mask(q_segment_ids, kv_segment_ids)
			if causal:
				span_q = start_q * block_q + jnp.arange(block_q)
				span_k = start_k * block_k + jnp.arange(block_k)
				causal_mask = span_q[:, None] >= span_k[None, :]
				mask = causal_mask if mask is None else jnp.logical_and(mask, causal_mask)
			# Apply mask to qk.
			qk = jnp.where(mask, qk, DEFAULT_MASK_VALUE)

		m_curr = qk.max(axis=-1)
		m_next = jnp.maximum(m_prev, m_curr)
		correction = jnp.exp(m_prev - m_next)
		l_prev_corr = correction * l_prev
		s_curr = jnp.exp(
			qk - m_next[:, None]
		)  # Use m_next instead of m_curr to avoid a correction on l_curr
		l_curr = s_curr.sum(axis=-1)
		l_next = l_prev_corr + l_curr
		l_next_rcp = 1.0 / l_next
		s_curr = s_curr * l_next_rcp[:, None]
		o_prev_corr = (l_prev_corr * l_next_rcp)[:, None] * o_prev
		v = pl.load(v_ref, (curr_k_slice, pl.dslice(block_d)))
		o_curr = pl.dot(s_curr.astype(v.dtype), v)

		o_next = o_prev_corr + o_curr
		return o_next, m_next, l_next

	if causal:
		# Ceildiv (`pl.cdiv` and `//` do not work due to type of start_q)
		upper_bound = lax.div(block_q * (start_q + 1) + block_k - 1, block_k)
	else:
		upper_bound = pl.cdiv(seq_len, block_k)  # type: ignore
	o, m_i, l_i = lax.fori_loop(0, upper_bound, body, (o, m_i, l_i))

	if residual_refs:
		l_ref, m_ref = residual_refs
		pl.store(l_ref, (curr_q_slice,), l_i)
		pl.store(m_ref, (curr_q_slice,), m_i)
	# Write output to dram.
	o = o.astype(o_ref.dtype)
	pl.store(o_ref, (curr_q_slice, pl.dslice(None)), o)


def segment_mask(
	q_segment_ids: jax.Array,
	kv_segment_ids: jax.Array,
):
	# [B, T, 1] or [T, 1]
	q_segment_ids = jnp.expand_dims(q_segment_ids, axis=-1)
	# [B, 1, S] or [1, S]
	if kv_segment_ids.ndim == 1:
		kv_segment_ids = jnp.expand_dims(kv_segment_ids, axis=0)
	else:
		kv_segment_ids = jnp.expand_dims(kv_segment_ids, axis=1)
	return jnp.equal(q_segment_ids, kv_segment_ids).astype(jnp.bool_)


@functools.partial(jax.custom_vjp, nondiff_argnums=[4, 5, 6, 7, 8, 9, 10, 11, 12, 13])
@functools.partial(
	jax.jit,
	static_argnames=[
		"sm_scale",
		"causal",
		"block_q",
		"block_k",
		"backward_pass_impl",
		"num_warps",
		"num_stages",
		"grid",
		"interpret",
		"debug",
	],
)
def mha(
	q,
	k,
	v,
	segment_ids: jnp.ndarray | None,
	sm_scale: float = 1.0,
	causal: bool = False,
	block_q: int = 128,
	block_k: int = 128,
	backward_pass_impl: str = "triton",
	num_warps: int | None = None,
	num_stages: int = 2,
	grid: tuple[int, ...] | None = None,
	interpret: bool = False,
	debug: bool = False,
):
	del backward_pass_impl
	batch_size, seq_len, num_heads, head_dim = q.shape
	block_q = min(block_q, seq_len)
	block_k = min(block_k, seq_len)
	# Heuristics.
	grid_ = grid
	if grid_ is None:
		grid_ = (pl.cdiv(seq_len, block_q), batch_size, num_heads)

	num_warps_ = num_warps
	if num_warps_ is None:
		num_warps_ = 4 if head_dim <= 64 else 8
	kernel = functools.partial(
		mha_forward_kernel,
		num_heads=num_heads,
		sm_scale=sm_scale,
		block_q=block_q,
		block_k=block_k,
		block_d=head_dim,
		causal=causal,
	)

	in_specs = [
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
	]
	in_specs.append(
		None  # type: ignore[arg-type]
		if segment_ids is None
		else pl.BlockSpec((None, seq_len), lambda _, j, k: (j, 0))
	)
	out_shape = jax.ShapeDtypeStruct(shape=q.shape, dtype=q.dtype)
	return pl.pallas_call(
		kernel,
		grid=grid_,
		in_specs=in_specs,
		out_specs=pl.BlockSpec(
			(None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)
		),
		compiler_params=dict(triton=dict(num_warps=num_warps_, num_stages=num_stages)),
		out_shape=out_shape,
		debug=debug,
		interpret=interpret,
		name="mha_forward",
	)(q, k, v, segment_ids)


def _mha_forward(
	q,
	k,
	v,
	segment_ids: jax.Array | None,
	sm_scale: float,
	causal: bool,
	block_q: int,
	block_k: int,
	backward_pass_impl: str,
	num_warps: int | None,
	num_stages: int,
	grid: Any,
	interpret: bool,
	debug: bool,
):
	del backward_pass_impl
	batch_size, seq_len, num_heads, head_dim = q.shape
	block_q = min(block_q, seq_len)
	block_k = min(block_k, seq_len)
	# Heuristics.
	grid_ = grid
	if grid_ is None:
		grid_ = (pl.cdiv(seq_len, block_q), batch_size, num_heads)

	num_warps_ = num_warps
	if num_warps_ is None:
		num_warps_ = 4 if head_dim <= 64 else 8
	kernel = functools.partial(
		mha_forward_kernel,
		num_heads=num_heads,
		sm_scale=sm_scale,
		causal=causal,
		block_q=block_q,
		block_k=block_k,
		block_d=head_dim,
	)
	out_shape = [
		jax.ShapeDtypeStruct(shape=q.shape, dtype=q.dtype),  # out
		jax.ShapeDtypeStruct(
			shape=(batch_size, num_heads, seq_len),  # l
			dtype=jnp.float32,
		),
		jax.ShapeDtypeStruct(
			shape=(batch_size, num_heads, seq_len),  # m
			dtype=jnp.float32,
		),
	]
	in_specs = [
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
		pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
	]
	in_specs.append(
		None  # type: ignore[arg-type]
		if segment_ids is None
		else pl.BlockSpec((None, seq_len), lambda _, j, k: (j, 0))
	)
	out, l, m = pl.pallas_call(
		kernel,
		grid=grid_,
		in_specs=in_specs,
		out_specs=[
			pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, None, seq_len), lambda _, j, k: (j, k, 0)),
			pl.BlockSpec((None, None, seq_len), lambda _, j, k: (j, k, 0)),
		],
		compiler_params=dict(triton=dict(num_warps=num_warps_, num_stages=num_stages)),
		out_shape=out_shape,
		debug=debug,
		interpret=interpret,
		name="mha_forward",
	)(q, k, v, segment_ids)
	return out, (q, k, v, segment_ids, out, l, m)


def _preprocess_backward_kernel(
	out_ref, dout_ref, l_ref, new_dout_ref, delta_ref, *, block_q: int
):
	pid_m = pl.program_id(0)

	off_m = pl.ds(pid_m * block_q, block_q)
	# load
	o = pl.load(out_ref, (off_m, slice(None))).astype(jnp.float32)
	do = pl.load(dout_ref, (off_m, slice(None))).astype(jnp.float32)
	denom = pl.load(l_ref, (off_m,)).astype(jnp.float32)
	# compute
	do = do / denom[:, None]
	delta = jnp.sum(o * do, axis=1)
	# write-back
	pl.store(new_dout_ref, (off_m, slice(None)), do.astype(new_dout_ref.dtype))
	pl.store(delta_ref, (off_m,), delta.astype(delta_ref.dtype))


@jax.named_scope("preprocess_backward")
def _preprocess_backward(out, do, l, block_q: int, debug: bool, interpret: bool):
	batch_size, seq_len, num_heads, head_dim = out.shape
	out_shape = [
		jax.ShapeDtypeStruct(do.shape, do.dtype),
		jax.ShapeDtypeStruct(l.shape, l.dtype),
	]
	do_scaled, delta = pl.pallas_call(
		functools.partial(_preprocess_backward_kernel, block_q=block_q),
		grid=(pl.cdiv(seq_len, block_q), batch_size, num_heads),
		in_specs=[
			pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, None, seq_len), lambda _, j, k: (j, k, 0)),
		],
		out_specs=[
			pl.BlockSpec((None, seq_len, None, head_dim), lambda _, j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, None, seq_len), lambda _, j, k: (j, k, 0)),
		],
		compiler_params=dict(triton=dict(num_warps=4, num_stages=3)),
		out_shape=out_shape,
		debug=debug,
		interpret=interpret,
		name="mha_preprocess_backward",
	)(out, do, l)
	return do_scaled, delta


def mha_backward_kernel(
	# Inputs
	q_ref,
	k_ref,
	v_ref,
	segment_ids_ref: jax.Array | None,
	out_ref,
	do_scaled_ref,
	l_ref,
	m_ref,
	delta_ref,
	_,
	# Outputs
	dq_ref,
	dk_ref,
	dv_ref,
	*,
	sm_scale: float,
	causal: bool,
	block_q: int,
	block_d: int,
	block_k: int,
):
	del out_ref, l_ref  # Not needed
	seq_len = q_ref.shape[0]

	def outer_loop(start_k, _):
		dv = jnp.zeros([block_k, block_d], dtype=jnp.float32)
		dk = jnp.zeros([block_k, block_d], dtype=jnp.float32)
		k = pl.load(k_ref, (pl.ds(start_k * block_k, block_k), slice(None)))
		v = pl.load(v_ref, (pl.ds(start_k * block_k, block_k), slice(None)))
		span_k = start_k * block_k + jnp.arange(block_k)
		kv_segment_ids = (
			None
			if segment_ids_ref is None
			else pl.load(segment_ids_ref, (pl.ds(start_k * block_k, block_k),))
		)

		def inner_loop(start_q, carry):
			dv, dk = carry
			q = pl.load(q_ref, (pl.ds(start_q * block_q, block_q), slice(None)))
			qk = pl.dot(q, k.T)
			qk = qk.astype(q_ref.dtype)
			qk = qk.astype(jnp.float32)
			if sm_scale != 1.0:
				qk *= sm_scale

			q_segment_ids = (
				None
				if segment_ids_ref is None
				else pl.load(segment_ids_ref, (pl.ds(start_q * block_q, block_q),))
			)

			if causal or segment_ids_ref is not None:
				mask = None
				if segment_ids_ref is not None:
					mask = segment_mask(q_segment_ids, kv_segment_ids)

				if causal:
					span_q = start_q * block_q + jnp.arange(block_q)
					causal_mask = span_q[:, None] >= span_k[None, :]
					mask = causal_mask if mask is None else jnp.logical_and(mask, causal_mask)
				qk = jnp.where(mask, qk, DEFAULT_MASK_VALUE)

			m = pl.load(m_ref, (pl.ds(start_q * block_q, block_q),))
			p = jnp.exp(qk - m[:, None])
			do = pl.load(do_scaled_ref, (pl.ds(start_q * block_q, block_q), slice(None)))
			dv = dv + pl.dot(p.astype(do.dtype).T, do)
			di = pl.load(delta_ref, (pl.ds(start_q * block_q, block_q),))
			dp = jnp.zeros((block_q, block_k), dtype=jnp.float32) - di[:, None]
			dp = dp + pl.dot(do, v.T)
			ds = p * dp
			if sm_scale != 1.0:
				ds = ds * sm_scale
			dk = dk + pl.dot(ds.astype(q_ref.dtype).T, q)
			dq = pl.load(
				dq_ref,
				(pl.ds(start_q * block_q, block_q), slice(None)),
				eviction_policy="evict_last",
			)
			dq = dq + pl.dot(ds.astype(k.dtype), k).astype(dq.dtype)
			pl.store(
				dq_ref,
				(pl.ds(start_q * block_q, block_q), slice(None)),
				dq,
				eviction_policy="evict_last",
			)
			return dv, dk

		if causal:
			lower_bound = lax.div(start_k * block_k, block_q)
		else:
			lower_bound = 0
		dv, dk = lax.fori_loop(lower_bound, pl.cdiv(seq_len, block_q), inner_loop, (dv, dk))
		pl.store(
			dv_ref, (pl.ds(start_k * block_k, block_k), slice(None)), dv.astype(dv_ref.dtype)
		)
		pl.store(
			dk_ref, (pl.ds(start_k * block_k, block_k), slice(None)), dk.astype(dk_ref.dtype)
		)

	lax.fori_loop(0, pl.cdiv(seq_len, block_k), outer_loop, None)


def _mha_backward(
	sm_scale: float,
	causal: bool,
	block_q: int,
	block_k: int,
	backward_pass_impl: str,
	num_warps: int | None,
	num_stages: int,
	grid: Any,
	interpret: bool,
	debug: bool,
	res,
	do,
):
	del num_warps, num_stages, grid
	q, k, v, segment_ids, out, l, m = res

	if backward_pass_impl == "xla":
		return jax.vjp(
			functools.partial(mha_reference, sm_scale=sm_scale, causal=causal),
			q,
			k,
			v,
			segment_ids,
		)[1](do)
	elif backward_pass_impl == "triton":
		batch_size, seq_len, num_heads, head_dim = q.shape
		block_q = min(block_q, seq_len)
		block_k = min(block_k, seq_len)
		do_scaled, delta = _preprocess_backward(out, do, l, block_q, debug, interpret)
		# We accumulate into dq so we need to initialize it to zeros.
		dq = jnp.zeros(q.shape, jnp.float32)
		out_shapes = [
			jax.ShapeDtypeStruct(dq.shape, dq.dtype),
			jax.ShapeDtypeStruct(k.shape, k.dtype),
			jax.ShapeDtypeStruct(v.shape, v.dtype),
		]

		in_specs = [
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			pl.BlockSpec((None, None, seq_len), lambda j, k: (j, k, 0)),
			pl.BlockSpec((None, None, seq_len), lambda j, k: (j, k, 0)),
			pl.BlockSpec((None, None, seq_len), lambda j, k: (j, k, 0)),
			pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
		]
		if segment_ids is None:
			in_specs.insert(3, None)  # type: ignore[arg-type]
			input_output_aliases = {8: 0}
		else:
			in_specs.insert(3, pl.BlockSpec((None, seq_len), lambda j, k: (j, 0)))
			input_output_aliases = {9: 0}
		grid = (batch_size, num_heads)
		num_warps = 8
		dq, dk, dv = pl.pallas_call(
			functools.partial(
				mha_backward_kernel,
				block_q=block_q,
				block_d=head_dim,
				block_k=block_k,
				sm_scale=sm_scale,
				causal=causal,
			),
			grid=grid,
			out_shape=out_shapes,
			in_specs=in_specs,
			out_specs=[
				pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
				pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
				pl.BlockSpec((None, seq_len, None, head_dim), lambda j, k: (j, 0, k, 0)),
			],
			name="mha_backward",
			debug=debug,
			interpret=interpret,
			compiler_params=dict(triton=dict(num_warps=num_warps, num_stages=1)),
			input_output_aliases=input_output_aliases,
		)(q, k, v, segment_ids, out, do_scaled, l, m, delta, dq)
	else:
		raise ValueError(f"Invalid backward pass implementation: {backward_pass_impl}")
	return dq.astype(q.dtype), dk, dv, None


mha.defvjp(_mha_forward, _mha_backward)


@functools.partial(jax.jit, static_argnames=["sm_scale", "causal"])
def mha_reference(
	q,
	k,
	v,
	segment_ids: jnp.ndarray | None,
	sm_scale=1.0,
	causal: bool = False,
):
	q_seq_len = q.shape[1]
	kv_seq_len = k.shape[1]
	logits = jnp.einsum("bqhc,bkhc->bhqk", q, k).astype(jnp.float32)
	mask = None
	if segment_ids is not None:
		mask = jnp.expand_dims(segment_mask(segment_ids, segment_ids), 1)
		mask = jnp.broadcast_to(mask, logits.shape)
	if causal:
		causal_mask = jnp.tril(jnp.ones((1, 1, q_seq_len, kv_seq_len), dtype=bool))
		causal_mask = jnp.broadcast_to(causal_mask, logits.shape)
		mask = causal_mask if mask is None else jnp.logical_and(mask, causal_mask)
	logits = logits if mask is None else jnp.where(mask, logits, float("-inf"))
	weights = jax.nn.softmax(logits * sm_scale).astype(q.dtype)
	return jnp.einsum("bhqk,bkhc->bqhc", weights, v)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flax layer for AQT injection."""

import copy
import enum
import functools
from typing import Iterable
from typing import Optional
from fjformer.bit_quantization import q_dot_general
from fjformer.bit_quantization import calibration
from fjformer.bit_quantization import config
from fjformer.bit_quantization import int_numerics
from fjformer.bit_quantization import no_numerics
import flax
import flax.linen as nn
import jax
from jax._src.numpy import lax_numpy
import jax.numpy as jnp


class QuantMode(enum.Enum):
	TRAIN = 1
	CONVERT = 2
	SERVE = 3


class Freezer(nn.Module):
	"""Identity function that can freeze its input.

	On default it is an identity function that saves the input in a variable.
	In 'use_frozen=True' mode, ignores the input and returns the frozen value. It
	is usefult to implement 'constant folding' and put quantized weights and
	scales in the checkpoint for serving.
	"""

	quant_collection: str
	quant_mode: QuantMode
	q_shape: Iterable[int]
	q_init: nn.initializers.Initializer
	s_shape: Iterable[int]
	s_init: nn.initializers.Initializer

	@nn.compact
	def __call__(
		self, inputs: Optional[q_dot_general.QTensor]
	) -> Optional[q_dot_general.QTensor]:
		collection = self.quant_collection
		if inputs is None:  # getter mode
			if self.quant_mode == QuantMode.TRAIN:
				return inputs
			elif self.quant_mode == QuantMode.CONVERT:
				return inputs
			elif self.quant_mode == QuantMode.SERVE:
				# We could have created one self.variable whose value is a QTensor,
				# but this would complicate the init function, which could potentially
				# be used by adding metadata such as sharding axises, etc.
				qvalue = self.variable(collection, "value", self.q_init, self.q_shape)
				scale = self.variable(collection, "scale", self.s_init, self.s_shape)
				return q_dot_general.QTensor(qvalue.value, scale.value)
			else:
				raise AssertionError("Unknown quant mode.") from None
		else:  # setter mode
			if self.quant_mode == QuantMode.TRAIN:
				pass
			elif self.quant_mode == QuantMode.CONVERT:
				qvalue = self.variable(collection, "value", self.q_init, self.q_shape)
				scale = self.variable(collection, "scale", self.s_init, self.s_shape)
				qvalue.value = inputs.qvalue
				scale.value = inputs.qvalue_scale_t
			elif self.quant_mode == QuantMode.SERVE:
				pass
			else:
				raise AssertionError("Unknown quant mode.") from None
			return None


class QDotGeneral(nn.Module):
	"""A layer that can be injected into flax.nn.Dense, etc."""

	cfg: Optional[config.DotGeneral] = None
	prng_name: Optional[str] = "params"

	lhs_quant_mode: QuantMode = QuantMode.TRAIN
	lhs_init: nn.initializers.Initializer = jnp.zeros
	lhs_scale_init: nn.initializers.Initializer = jnp.zeros
	lhs_var_name: str = "qlhs"

	rhs_quant_mode: QuantMode = QuantMode.TRAIN
	rhs_init: nn.initializers.Initializer = jnp.zeros
	rhs_scale_init: nn.initializers.Initializer = jnp.zeros
	rhs_var_name: str = "qrhs"

	# If you want use 'params' make sure that there is another mechanism to hide
	# these variables from the optimizer.
	quant_collection: str = "aqt"

	def make_aqt_dg(
		self,
		lhs_shape,
		rhs_shape,
		dimension_numbers: tuple[Iterable[int], Iterable[int]],
	):
		lhs_scale_shape = list(lhs_shape)
		rhs_scale_shape = list(rhs_shape)
		(contr, _) = dimension_numbers
		for li, ri in zip(*contr):  # noqa:B905
			lhs_scale_shape[li] = 1
			rhs_scale_shape[ri] = 1
		lhs_scale = q_dot_general._lhs_scale_transpose(  # pylint: disable=protected-access
			jnp.zeros(lhs_scale_shape), dimension_numbers, lhs_shape, rhs_shape
		)
		assert lhs_scale is not None
		lhs_scale_shape = lhs_scale.shape
		rhs_scale = q_dot_general._rhs_scale_transpose(  # pylint: disable=protected-access
			jnp.zeros(rhs_scale_shape), dimension_numbers, lhs_shape, rhs_shape
		)
		assert rhs_scale is not None
		rhs_scale_shape = rhs_scale.shape

		cfg = copy.deepcopy(self.cfg)
		if cfg is not None:
			rhs_qm = self.rhs_quant_mode
			lhs_qm = self.lhs_quant_mode

			msg = "The only function that is setting preprocess can be QQuantized."
			assert cfg.fwd.rhs.preprocess is None, msg
			assert cfg.fwd.lhs.preprocess is None, msg
			cfg.fwd.lhs.preprocess = Freezer(
				name=self.lhs_var_name,
				quant_mode=lhs_qm,
				q_shape=lhs_shape,
				q_init=self.lhs_init,
				s_shape=lhs_scale_shape,
				s_init=self.lhs_scale_init,
				quant_collection=self.quant_collection,
			)
			cfg.fwd.rhs.preprocess = Freezer(
				name=self.rhs_var_name,
				quant_mode=rhs_qm,
				q_shape=rhs_shape,
				q_init=self.rhs_init,
				s_shape=rhs_scale_shape,
				s_init=self.rhs_scale_init,
				quant_collection=self.quant_collection,
			)
		key = self.make_rng(self.prng_name) if self.prng_name is not None else None
		context = q_dot_general.Context(key=key, train_step=None)
		aqt_dg = q_dot_general.make_dot_general(cfg)
		aqt_dg = functools.partial(aqt_dg, context=context)
		return aqt_dg

	@nn.compact
	def __call__(
		self,
		lhs,
		rhs,
		dimension_numbers,
		precision,
		preferred_element_type=None,
	):
		aqt_dg = self.make_aqt_dg(lhs.shape, rhs.shape, dimension_numbers)
		return aqt_dg(
			lhs,
			rhs,
			dimension_numbers,
			precision,
			preferred_element_type=preferred_element_type,
		)


class QEinsum(flax.struct.PyTreeNode):
	"""Quantized Einsum class for model injection."""

	cfg: Optional[config.DotGeneral] = None
	prng_name: Optional[str] = "params"

	lhs_quant_mode: QuantMode = QuantMode.TRAIN
	lhs_init: nn.initializers.Initializer = jnp.zeros
	lhs_scale_init: nn.initializers.Initializer = jnp.zeros
	lhs_var_name: str = "qlhs"

	rhs_quant_mode: QuantMode = QuantMode.TRAIN
	rhs_init: nn.initializers.Initializer = jnp.zeros
	rhs_scale_init: nn.initializers.Initializer = jnp.zeros
	rhs_var_name: str = "qrhs"

	# If you want use 'params' make sure that there is another mechanism to hide
	# these variables from the optimizer.
	quant_collection: str = "aqt"

	def __call__(self, eqn, lhs_g, rhs_g):
		def einsum(lhs_l, rhs_l, dg=jax.lax.dot_general):
			operands, contractions = lax_numpy._default_poly_einsum_handler(  # pylint: disable=protected-access
				eqn, lhs_l, rhs_l, einsum_call=True, use_blas=True, optimize="optimal"
			)
			contractions = tuple((a, frozenset(b), c) for a, b, c, *_ in contractions)
			return jax.named_call(lax_numpy._einsum, name=eqn)(  # pylint: disable=protected-access
				operands,
				contractions,
				precision=None,
				preferred_element_type=None,
				_dot_general=dg,
			)

		# yes_swap = whether einsum swaps [lhs,rhs] when passing them to dot_general
		a = jax.make_jaxpr(einsum)(lhs_g, rhs_g)
		[lhs_g_id, rhs_g_id] = a.eqns[0].invars
		[lhs_l_id, rhs_l_id] = a.jaxpr.invars
		not_swap = lhs_g_id == lhs_l_id and rhs_g_id == rhs_l_id
		yes_swap = lhs_g_id == rhs_l_id and rhs_g_id == lhs_l_id
		assert not_swap != yes_swap

		cfg = copy.deepcopy(self.cfg)
		prng_name = self.prng_name

		lhs_quant_mode = self.lhs_quant_mode
		lhs_init = self.lhs_init
		lhs_scale_init = self.lhs_scale_init
		lhs_var_name = self.lhs_var_name

		rhs_quant_mode = self.rhs_quant_mode
		rhs_init = self.rhs_init
		rhs_scale_init = self.rhs_scale_init
		rhs_var_name = self.rhs_var_name

		quant_collection = self.quant_collection

		if yes_swap:
			if cfg is not None:
				cfg.fwd.lhs, cfg.fwd.rhs = cfg.fwd.rhs, cfg.fwd.lhs
				cfg.dlhs, cfg.drhs = cfg.drhs, cfg.dlhs
			lhs_quant_mode, rhs_quant_mode = rhs_quant_mode, lhs_quant_mode
			lhs_init, rhs_init = rhs_init, lhs_init
			lhs_scale_init, rhs_scale_init = rhs_scale_init, lhs_scale_init
			lhs_var_name, rhs_var_name = rhs_var_name, lhs_var_name

		aqt_dg = QDotGeneral(
			cfg=cfg,
			prng_name=prng_name,
			lhs_quant_mode=lhs_quant_mode,
			lhs_init=lhs_init,
			lhs_scale_init=lhs_scale_init,
			lhs_var_name=lhs_var_name,
			rhs_quant_mode=rhs_quant_mode,
			rhs_init=rhs_init,
			rhs_scale_init=rhs_scale_init,
			rhs_var_name=rhs_var_name,
			quant_collection=quant_collection,
		)
		return einsum(lhs_g, rhs_g, aqt_dg)


def config_v4(
	*,
	fwd_bits: Optional[int] = 8,
	dlhs_bits: Optional[int] = 8,
	drhs_bits: Optional[int] = None,
	# The dummy static bound flag is for performance benchmarking.
	use_dummy_static_bound: bool = False,
	rng_type: str = "jax.uniform",  # 'custom-1'
	dlhs_local_aqt: Optional[config.LocalQ] = None,
	drhs_local_aqt: Optional[config.LocalQ] = None,
	fwd_accumulator_dtype: ... = jnp.int32,
	dlhs_accumulator_dtype: ... = jnp.int32,
	drhs_accumulator_dtype: ... = None,
) -> config.DotGeneral:
	"""
	The config_v4 function is a helper function that creates a DotGeneral config
	object. It takes in the following arguments:
	- fwd_bits: The number of bits to use for forward pass quantization. If None, no quantization will be used.
	Defaults to 8 bits.
	- dlhs_bits: The number of bits to use for left hand side gradient quantization (i.e., the weights). If None,
	 no quantization will be used. Defaults to 8 bits..
	- drhs_bits: The number of bits to use for right hand side gradient quanitzation

	:param *: Indicate that the function accepts a variable number of arguments
	:param fwd_bits: Optional[int]: Set the number of bits for the forward pass
	:param dlhs_bits: Optional[int]: Set the number of bits used for quantization
	:param drhs_bits: Optional[int]: Set the number of bits for the right hand side
	:param use_dummy_static_bound: bool: Set the static bound to 1
	:param rng_type: str: Set the type of random number generator
	:param dlhs_local_aqt: Optional[config.LocalQ]: Set the local quantization parameters for the lhs gradient
	:param drhs_local_aqt: Optional[config.LocalQ]: Set the local quantization parameters for the drhs tensor
	:param fwd_accumulator_dtype: ...: Set the accumulator dtype for forward pass
	:param dlhs_accumulator_dtype: ...: Determine the dtype of the accumulator in
	:param drhs_accumulator_dtype: ...: Determine the dtype of the accumulator used in q_dot_general
	:param : Determine the number of bits used for quantization
	:return: A config
	"""

	def tensor_config(bits: Optional[int]) -> config.Tensor:
		assert bits is None or bits >= 2, "Need at least 2 bits."
		if bits is None:
			numerics = no_numerics.NoNumerics()
		else:
			numerics = int_numerics.IntNumerics(
				bits=bits,
				preserve_zero=True,
				preserve_max_val=False,
				clip=True,
				round=True,
				noise_fn=None,
				clip_gradient=False,  # Can be False when using abs-max scaling.
				dtype=jnp.int8 if 2 <= bits <= 8 else None,
			)

		return config.Tensor(
			numerics=numerics,
			calib_shared_axes=None,
			scale_stop_grad=True,
			calibration=calibration.AbsMaxCalibration(),
			po2_scale=False,
			use_fake_quant=False,
			# dtype_x=dtype,
			use_fwd_quant=None,
			preprocess=None,
		)

	def dg_raw_config(lhs_bits, rhs_bits, local_aqt=None) -> config.DotGeneralRaw:
		lhs_cfg = tensor_config(lhs_bits)
		rhs_cfg = tensor_config(rhs_bits)
		if (
			True  # Just to format lines below
			and lhs_bits is not None
			and rhs_bits is not None
			and lhs_bits <= 8
			and rhs_bits <= 8
		):
			dg_accumulator_dtype = jnp.int32
		else:
			# None determines the dtype on the fly in q_dot_general
			dg_accumulator_dtype = None

		return config.DotGeneralRaw(
			lhs=lhs_cfg,
			rhs=rhs_cfg,
			dg_accumulator_dtype=dg_accumulator_dtype,
			local_aqt=local_aqt,
		)

	cfg = config.DotGeneral(
		fwd=dg_raw_config(fwd_bits, fwd_bits),
		dlhs=dg_raw_config(dlhs_bits, dlhs_bits, local_aqt=dlhs_local_aqt),
		drhs=dg_raw_config(drhs_bits, drhs_bits, local_aqt=drhs_local_aqt),
	)

	cfg.dlhs.rhs.use_fwd_quant = False
	cfg.drhs.rhs.use_fwd_quant = False

	# Typically we have (but I don't know if it is guraranteed):
	# - vjp_lhs_stochastic_rounding is referring to the gradient and
	# - vjp_rhs_stochastic_rounding is referring to the activations/weights.
	config.set_stochastic_rounding(
		cfg,
		vjp_lhs_stochastic_rounding=True,
		vjp_rhs_stochastic_rounding=False,
		implementation=rng_type,
	)

	if use_dummy_static_bound:
		config.set_static_bound(cfg, 1.0)

	config.set_accumulator_dtype(
		cfg,
		fwd_dtype=fwd_accumulator_dtype,
		dlhs_dtype=dlhs_accumulator_dtype,
		drhs_dtype=drhs_accumulator_dtype,
	)

	return cfg
